//
// Implementation of TwitterFox main component
//
// Copyright (c) 2009 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var Cc = Components.classes;
var Ci = Components.interfaces;
var TWITTERFOX_UUID = "twitternotifier@naan.net";
var CLASS_ID = Components.ID("3c4bfeef-c936-0d98-8fb9-b9b28e4f28ed");
var CLASS_NAME = "Twitter Notifier";
var CONTRACT_ID = "@naan.net/twitternotifier;1";

var MAX_STORED_MESSAGES = 40;

var TWEET_FRIENDS   = 0;
var TWEET_MENTIONS  = 1;
var TWEET_MESSAGES  = 2;
var TWEET_FAVORITES = 3;

var TWEET_TYPES = {
  "timeline": 0,
  "replies":  1,
  "messages": 2,
};

//
// cached sql statements
// 
var find_user_stmt = null;
var insert_user_stmt = null;
var insert_status_stmt = null;
var insert_dm_stmt = null;
var insert_timeline_stmt = null;

var db = null;

var TWITTER_API_URL = "https://twitter.com/";
var NETWORK_TIMEOUT_TIME = 120;
var APP_NAME = "TwitterFox";

function dumpall(obj) 
{
  for (var i in obj) {
    if (typeof obj[i] == 'object') {
      dump(i + ":\n");
      this.dumpall(obj[i]);
    }
    else if (typeof obj[i] != 'function') {
      dump(i + ":" + obj[i] + "\n");
    } 
  }
}

/////////////////////////////////////////////////////////////////////////
//
// Twitter status model
//
function User(json)
{
  for (var i in json) {
    this[i] = json[i];
  }
}

User.prototype = {
  insertIntoDB: function() {
    if (insert_user_stmt == null) {
      insert_user_stmt = db.prepare("REPLACE INTO users VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
    }
    var stmt = insert_user_stmt;
    stmt.bindInt64Parameter(  0, this.id);
    stmt.bindStringParameter( 1, this.name);
    stmt.bindStringParameter( 2, this.screen_name);
    stmt.bindStringParameter( 3, this.location);
    stmt.bindStringParameter( 4, this.description);
    stmt.bindStringParameter( 5, this.url);
    stmt.bindInt32Parameter(  6, this.followers_count);
    stmt.bindInt32Parameter(  7, this.friends_count);
    stmt.bindInt32Parameter(  8, this.favorites_count);
    stmt.bindInt32Parameter(  9, this.statuses_count);
    stmt.bindStringParameter(10, this.profile_image_url);
    stmt.bindInt32Parameter( 11, this.protected);
    stmt.bindInt64Parameter( 12, this.updated_at);
    try {
      stmt.execute();
    }
    catch (e) {
      db.logError("Error on insert user");
    }
    stmt.reset();
  },
 
  updateDB: function() {
  },

  findById: function(id) {
    this.id = id;

    if (find_user_stmt == null) {
      find_user_stmt = db.prepare("SELECT * FROM users WHERE user_id = ?");
    }
    var stmt = find_user_stmt;
    stmt.bindInt64Parameter( 0, id);

    while (stmt.executeStep()) {
      // from statuses table
      this.id                = stmt.getInt32(0);
      this.name              = stmt.getString(1);
      this.screen_name       = stmt.getString(2);
      this.location          = stmt.getString(3);
      this.description       = stmt.getString(4);
      this.url               = stmt.getString(5);
      this.followers_count   = stmt.getInt32(6);
      this.friends_count     = stmt.getInt32(7);
      this.favorites_count   = stmt.getInt32(8);
      this.statuses_count    = stmt.getInt32(9);
      this.profile_image_url = stmt.getString(10);
      this.protected         = stmt.getInt32(11);
      this.updated_at        = stmt.getInt64(12);
    }
    stmt.reset();
  }
};

function Status(json, type)
{
  for (var i in json) {
    this[i] = json[i];
  }
  this.type = type;
}

Status.prototype = {
  insertIntoDB: function(recipient_id) {
    var stmt;
    if (insert_status_stmt == null) {
      insert_status_stmt = db.prepare("REPLACE INTO statuses VALUES (?, ?, ?, ?, ?, ?, ?, ?)");
    }
    stmt = insert_status_stmt;
    stmt.bindInt64Parameter( 0, this.id);
    stmt.bindInt32Parameter( 1, this.user.id);
    stmt.bindStringParameter(2, this.text);
    stmt.bindInt64Parameter( 3, new Date(this.created_at).getTime());
    stmt.bindStringParameter(4, this.source);
    stmt.bindInt64Parameter( 5, this.in_reply_to_status_id);
    stmt.bindInt32Parameter( 6, this.in_reply_to_user_id);
    stmt.bindStringParameter(7, this.in_reply_to_screen_name);
    try {
      stmt.execute();
    }
    catch (e) {
      db.logError("Error on insert status");
    }
    stmt.reset();

    if (insert_timeline_stmt == null) {
      insert_timeline_stmt = db.prepare("REPLACE INTO timeline VALUES (?, ?, ?, ?)");
    }
    stmt = insert_timeline_stmt;
    stmt.bindInt64Parameter( 0, recipient_id);
    stmt.bindInt32Parameter( 1, TWEET_TYPES[this.type]);
    stmt.bindStringParameter(2, this.id);
    stmt.bindInt32Parameter( 3, this.favorited);
    try {
      stmt.execute();
    }
    catch (e) {
      db.logError("Error on insert timeline");
      stmt.reset();
      return false;
    }
    stmt.reset();

    this.user = new User(this.user);
    this.user.insertIntoDB();
    return true;
  },

  initWithStatement: function(stmt) {
    // from statuses table
    this.id                      = stmt.getInt64(0);
    this.user_id                 = stmt.getInt32(1);
    this.text                    = stmt.getString(2);
    this.created_at              = stmt.getInt64(3);
    this.source                  = stmt.getString(4);
    this.in_reply_to_status_id   = stmt.getInt64(5);
    this.in_reply_to_user_id     = stmt.getInt32(6);
    this.in_reply_to_screen_name = stmt.getString(7);
    // from timeline table
    this.recipient_id            = stmt.getInt32(8);
    this.type                    = TWEET_TYPES[stmt.getInt32(9)];
    //this.id                      = stmt.getInt32(10);
    this.favorited               = stmt.getInt32(11) ? true : false;

    this.user = new User();
    this.user.findById(this.user_id);

  }
};

//
// Twitter direct message model
//
function DirectMessage(json)
{
  for (var i in json) {
    this[i] = json[i];
  }
}


DirectMessage.prototype = {
  insertIntoDB: function() {
    var stmt;
    if (insert_dm_stmt == null) {
      insert_dm_stmt = db.prepare("REPLACE INTO direct_messages VALUES (?, ?, ?, ?, ?, ?, ?)");
    }
    stmt = insert_dm_stmt;
    stmt.bindInt64Parameter( 0, this.id);
    stmt.bindInt32Parameter( 1, this.sender_id);
    stmt.bindInt32Parameter( 2, this.recipient_id);
    stmt.bindStringParameter(3, this.text);
    stmt.bindInt64Parameter(4, new Date(this.created_at).getTime());
    stmt.bindStringParameter(5, this.sender_screen_name);
    stmt.bindStringParameter(6, this.recipient__screen_name);
    try {
      stmt.execute();
    }
    catch (e) {
      db.logError("Error on insert status");
      stmt.reset();
      return false;
    }
    stmt.reset();

    this.sender = new User(this.sender);
    this.sender.insertIntoDB();
    this.recipient = new User(this.recipient);
    this.recipient.insertIntoDB();
    return true;
  }
};
    
/////////////////////////////////////////////////////////////////////////
//
// TwitterFox main component.
//
function TwitterNotifier() {
  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);

  obs.addObserver(this, "xpcom-shutdown", false);
  obs.addObserver(this, "twitternotifier-command", false);

  // Setup sound notification
  this._sound = Cc["@mozilla.org/sound;1"].createInstance(Ci.nsISound);
}

// This is the implementation of your component.
TwitterNotifier.prototype = {
  _timer: null,
  _accounts: [],
  _newMessages: [],
  _unread: {"timeline":[], "replies":[], "messages":[]},
  _tinyURLs: [],
  _decodedTinyURLs: [],
  _getDirectMessage: true,

  //
  // commands
  //
  createRequest: function(func, param, isPost) {
    var ns = func.split(".");
    var requestURL = "";
    var requestParam = "";

    var request = Cc["@naan.net/twitterfox-http-request;1"].createInstance(Ci.nsITwitterFoxHttpRequest);

    if (ns.length > 1) {
      request.callback = ns[0] + "_" + ns[1];
      requestURL       = ns.join("/");
    }
    else {
      request.callback = func;
      requestURL       = func;
    }

    requestURL = TWITTER_API_URL + requestURL + ".json";

    if (!param) param = {};

    for (var attr in param) {
      if (requestParam) requestParam += "&";
      requestParam += attr + "=" + encodeURIComponent(param[attr]);
    }

    // Setup callbacks
    //
    request.setBasicAuth(this._user, this._pass);
    var target = this;
    request.onload    = function(p) {target.onLoad(request, param);}
    request.onerror   = function(p) {target.onError(request, param);}
    request.ontimeout = function(p) {target.onTimeout(request, param);}
    
    // send async request
    if (isPost) {
      request.setURL(requestURL);
      request.setPostData(requestParam);
    }
    else {
      request.setURL(requestURL + '?' + requestParam);
    }

    request.asyncOpen();

    return request;
  },

  get: function(method, params) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }
    this._req = this.createRequest(method, params, false);
  },

  post: function(method, params) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }
    this.createRequest(method, params, true);
  },

  restoreStatusesFromDB: function(recipient_id, type, count) {
    var results = [];

    var sql = "SELECT * FROM statuses, timeline WHERE timeline.recipient_id = ? AND type = ? AND statuses.id = timeline.id ORDER BY id DESC LIMIT ?";
    var stmt = db.prepare(sql);

    if (!count) count = 20;
    stmt.bindInt32Parameter(0, recipient_id);
    stmt.bindInt32Parameter(1, TWEET_TYPES[type]);
    stmt.bindInt32Parameter(2, count);

    try {
      while (stmt.executeStep()) {
        var status = new Status(null, type);

        status.initWithStatement(stmt);

        if (this._unread[type].indexOf(status.id) >= 0) {
          status.unread = true;
        }

        status.type = type;

        results.push(status);
      }
    }
    catch (e) {
      db.logError("Failed to restore " + type + " from local cache");
    }
    stmt.finalize();

    return results;
  },

  deleteStatusFromDB: function(status_id, recipient_id) {
    var stmt = db.prepare("DELETE FROM timeline WHERE id = ?1 AND recipient_id = ?2");
    stmt.bindInt64Parameter(0, status_id);
    stmt.bindInt32Parameter(1, recipient_id);
    try {
      stmt.execute();
    }
    catch (e) {
      db.log("Error on delete status");
    }
    stmt.finalize();
  },

  restoreMessagesFromDB: function(recipient_id, count) {
    var results = [];

    var sql = "SELECT * FROM direct_messages WHERE recipient_id = ? ORDER BY id DESC LIMIT ?";
    var stmt = db.prepare(sql);

    if (!count) count = 20;
    stmt.bindInt32Parameter(0, recipient_id);
    stmt.bindInt32Parameter(1, count);

    try {
      while (stmt.executeStep()) {
        var msg = new DirectMessage();
        // from statuses table
        msg.id                      = stmt.getInt64(0);
        msg.sender_id               = stmt.getInt32(1);
        msg.recipient__id           = stmt.getInt32(2);
        msg.text                    = stmt.getString(3);
        msg.created_at              = stmt.getInt64(4);

        if (this._unread["messages"].indexOf(msg.id) >= 0) {
          msg.unread = true;
        }

        msg.sender = new User();
        msg.sender.findById(msg.sender_id);
        msg.recipient = new User();
        msg.recipient.findById(msg.recipient_id);

        msg.type = "messages";

        results.push(msg);
      }
    }
    catch (e) {
      db.logError("Failed to restore direct messages from local cache");
    }
    stmt.finalize();

    return results;
  },

  deleteMessageFromDB: function(message_id, recipient_id) {
    var stmt = db.prepare("DELETE FROM direct_messages WHERE id = ?1 AND recipient_id = ?2");
    stmt.bindInt64Parameter(0, message_id);
    stmt.bindInt32Parameter(1, recipient_id);
    try {
      db.begin();
      stmt.execute();
      db.commit();
    }
    catch (e) {
      db.logError("Error on delete direct message");
      db.rollback();
    }
    stmt.finalize();
  },

  initSession: function(user) {

    if (db == null) {
      db = Cc["@naan.net/twitterfox-database;1"].getService(Ci.nsITwitterFoxDatabase);
      if (db.openDatabase() == 0) {
        dump("Re-create database");
        db.createDatabase();
        db.openDatabase();
      }
    }

    if (this._user == null) {
      if (this._timer) {
        this._timer.cancel();
      }

      this._user = user.user;
      this._pass = user.pass;

      this.updatePref();
      this.verifyCredentials();
    }
    else {
      // Send username and turn on twitterfox icon
      this.notifyStatus("updateUsername", this._accounts[this._user]);
    }
  },

  updatePref: function() {
    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);

    this._interval = pref.getIntPref("extensions.twitternotifier.interval") * 60 * 1000;
    // fail safe
    if (this._interval < 180 * 1000) {
      this._interval = 180 * 1000;
    }

    if (this._rateLimit) {
      this.setInterval();
    }
  },

  setNextTimer: function() {
    this._timer = this.setDelayTask(this._interval, "updateTweets", null);
  },

  verifyCredentials: function() {
    this.get("account.verify_credentials");
  },

  updateTweets: function() {
    if (this._user == null) return;

    if (this._timer) {
      this._timer.cancel();
    }

    /*
     * for debugging
     *
    data = this.readFile("test", "friends_timeline.json");
    var obj = JSON.fromString(data);
    this.statuses_friends_timeline(obj, null);
    return;
    */
    this._newMessages = [];
    this.get("statuses.friends_timeline", this.getLatestId("timeline"));
  },

  getLatestId: function(type) {

    var stmt;
    var recipient_id = this._accounts[this._user].id;
    if (type == "messages") {
      var stmt = db.prepare("SELECT id FROM direct_messages WHERE recipient_id = ? ORDER BY id DESC LIMIT 1");
      stmt.bindInt32Parameter(0, recipient_id);
    }
    else if (type == "replies") {
      var stmt = db.prepare("SELECT id FROM timeline WHERE recipient_id = ?1 AND timeline.type = ?2 ORDER BY id DESC LIMIT 1");
      stmt.bindInt32Parameter(0, recipient_id);
      stmt.bindInt32Parameter(1, TWEET_TYPES[type]);
    }
    else {
      var stmt = db.prepare("SELECT timeline.id FROM timeline,statuses WHERE timeline.id = statuses.id AND timeline.recipient_id = ?1 AND timeline.type = ?2 AND statuses.user_id != ?1 ORDER BY timeline.id DESC LIMIT 1");
      stmt.bindInt32Parameter(0, recipient_id);
      stmt.bindInt32Parameter(1, TWEET_TYPES[type]);
    }
    var ret = 0;
    
    try {
      stmt.executeStep();
      ret = stmt.getInt64(0);
    }
    catch (e) {
      db.logError("Failed to select " + type + " id");
    }

    if (ret) {
      return {since_id:ret, count:200};
    }
    else {
      return null;
    }
  },

  getTweets: function() {
    if (this._idle == STATE_ACTIVE) {
      this.get("statuses.friends_timeline", this.getLatestId('timeline'));
    }
    else {
      this._idle = STATE_SUSPEND;
    }
  },

  getReplies: function() {
    this.get("statuses.replies", this.getLatestId('replies'));
  },

  getDirectMessages: function() {
    if (this._getDirectMessage) {
      this.get("direct_messages", this.getLatestId('messages'));
    }
    else {
      this.updateTimeline();
    }
    this._getDirectMessage = !this._getDirectMessage;
  },

  getRecent: function(obj) {
    var type = obj.type;
    if (this._accounts[this._user]) {
      var ret = {};
      var count = this._unread[type].length;
      if (count < 20) count = 20;

      if (type == "messages") {
        ret.msgs = this.restoreMessagesFromDB(this._accounts[this._user].id, count);
      }
      else {
        ret.msgs = this.restoreStatusesFromDB(this._accounts[this._user].id, type, count);
      }
      ret.type = type;

      ret.unread = {};
      for (var i in this._unread) {
        ret.unread[i] = this._unread[i].length;
      }

      this.notifyStatus("openWindow", ret);
    }
  },

  getUnreadCount: function() {
    var unread = 0;
    for (var i in this._unread) {
      unread += this._unread[i].length;
    }
    return unread;
  },

  markRead: function(obj) {
    this._unread[obj.type] = [];
  },

  markAllRead: function() {
    var types = ['timeline', 'replies', 'messages'];
    for (var i in types) {
      this.markRead({type:types[i]});
    }
  },

  sendMessage: function(msg) {
    var arr = /^d\s+(\S+)\s+(.*)/(msg.status);
    if (arr && arr.length == 3) {
      this.post("direct_messages.new", {user: arr[1], text: arr[2]});
    }
    else {
      msg.source = APP_NAME;
      var status = {status:msg.status, source:APP_NAME};
      if (msg.inReplyTo) {
        status["in_reply_to_status_id"] = msg.inReplyTo;
      }
      this.post("statuses.update", status);
    }
  },

  setFavorite: function(msg) {
    this.post("favorites."  + msg.method + "." + msg.id);
  },

  deleteTweet: function(obj) {
    if (obj.type == 'timeline' || obj.type == 'replies') {
      this.post("statuses.destroy", {id:obj.id});
    }
    else {
      this.post("direct_messages.destroy", {id:obj.id});
    }
  },

  changeAccount: function(account) {

    // reset session
    this.reset();
    this.markAllRead();
    
    if (!this._accounts[account.user]) {
      this.initSession(account);
    }
    else {
      this._user = account.user;
      this._pass = account.pass;
      this.notifyStatus("updateUsername", {screen_name: this._accounts[account.user].screen_name,
            profile_image_url: this._accounts[account.user].profile_image_url});
      this.notifyStatus("accountChanged");
      this.updateTweets();
    }
  },

  reset: function() {
    if (this._timer) {
      this._timer.cancel();
    }

    if (this._req) {
      this._req.abort();
    }

    this._user = null;
    this._pass = null;
    this._hasError = false;
    this._getDirectMessage = true;
  },

  logout: function() {
    this.reset();
    this.notifyStatus("logout");
  },

  //
  // Private methods.
  //
  destroy: function(e) {
    if (this._timer) {
      this._timer.cancel();
      this._timer = null;
    }
    if (this._req) {
      this._req.abort();
    }

    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);

    obs.removeObserver(this, "xpcom-shutdown");
    obs.removeObserver(this, "twitternotifier-command");

    if (db) {
      db.close();
      db = null;
    }
 },

  retrieveTimeline: function(obj, req, method) {
    if (obj == null) return;

    try {
      // Added 'unread' flag if the message is new in any of stored messages
      this.log("Get " + obj.length + " " + method);
      db.begin();
      for (var i in obj) {

        var msg;
        var result;

        if (method == "messages") {
          msg = new DirectMessage(obj[i]);
          result = msg.insertIntoDB(this._accounts[this._user].id);
        }
        else {
          msg = new Status(obj[i], method);
          result = msg.insertIntoDB(this._accounts[this._user].id);
        }
        msg.type = method;

        // Do not insert to own message into new messages array
        if (msg.type != 'messages' && msg.user.id == this._accounts[this._user].id) continue;

        if (result) {
          if (method == 'replies') {
            // Do not insert duplicated replies into new messages array
            if (this._unread['timeline'].indexOf(msg.id) < 0) {
              this._unread[method].push(msg.id);
              this._newMessages.push(msg);
            }
          }
          else {
            this._unread[method].push(msg.id);
            this._newMessages.push(msg);
          }
        }
      }
      db.commit();
    }
    catch (e) {
      db.logError("Failed to insert " + method);
      db.rollback();
    }
  },

  updateTimeline: function() {
    if (this._newMessages.length) {
      // Play sound if this window is active
      this.playSound();
      this.notifyStatus("updateFriendsTimeline", this._newMessages);
    }
    else {
      if (!this._hasError) {
        this.notifyStatus("noUpdate");
      }
    }
    // reset error flag
    this._hasError = false;

    this._newMessages = [];
    this.setNextTimer();
  },

  playSound: function() {
    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.sound")) {

      try {
        var IOService = Cc['@mozilla.org/network/io-service;1'].getService(Ci.nsIIOService);
        var localFile = Cc['@mozilla.org/file/local;1'].createInstance(Ci.nsILocalFile);
        var url = pref.getCharPref("extensions.twitternotifier.soundFile");
      
        localFile.initWithPath(url);

        this._sound.play(IOService.newFileURI(localFile));
      }
      catch (e) {
      }
    }
  },

  //
  // Handling tinyURL
  //
  getDecodedTinyURL: function(link, messageId) {
    var TWITTERFOX_TINYURL = /[\s]*(http:\/\/(tinyurl.com|is.gd|bit.ly)\/\w+)[^\w\s\d]*/;
    if (!link.match(TWITTERFOX_TINYURL)) {
      return link;
    }

    if (this._decodedTinyURLs[link]) {
      return this._decodedTinyURLs[link];
    }
    else {
      var target = this;
      var req = Cc["@naan.net/twitterfox-http-request;1"].createInstance(Ci.nsITwitterFoxHttpRequest);

      req.setURL(link);
      req.setRedirectLimitation(0);
      req.onerror = function(e) {target.onErrorDecodeTinyURL(req, link, messageId);};
      req.asyncOpen();
      return link;
    }
  },

  onErrorDecodeTinyURL: function(req, url, messageId) {
    var loc = req.httpChannel().getResponseHeader("Location");
    this._decodedTinyURLs[url] = loc;
    this.notifyStatus("updateTinyURL", {url:loc, originalURL:url, messageId:messageId});
  },

  // Network handler
  //
  onLoad: function(req, param) {
    if (this._user == null) {
      this.reportError("No account information");
      return;
    }

    switch (Number(req.status)) {
    case 400:
      this.rateLimitExceeded(req);
      break;

    case 401:
      this._accounts[this._user] = null;
      this._user = null;
      this._pass = null;
      this.notifyStatus("authFail");
      break;

    case 200:
    case 304:
      var resp = null;
      if (!req.responseText.match(/^\s*$/)) {
        try {
          var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
          var resp = JSON.decode(req.responseText);
        }
        catch (e) {
          this.reportError("An error occurred while requesting " + req.__url);
          this.log("Response text: " + e.message);
          this.handleError(req, param, "Can't parse JSON. Twitter server responded with an error.");
        }
      }

      if (resp == null || (resp && resp.error)) {
        if (resp && resp.error) {
          this.notifyStatus("internalError", resp.error);
          this.setNextTimer();
        }
        else {
          this.handleError(req, param, "Twitter server responded with an error");
        }
      }
      else {
        this[req.callback](resp.statuses ? resp.statuses : resp, req);
      }
      break;

    case 403:
    case 404:
    case 406:
    case 500:
    case 502:
    case 503:
    default:
      this.handleError(req, param, "Twitter server responded with an error (" + req.status + ")");
      break;
 
    }
  },

  handleError: function(req, param, msg) {
    if (this[req.callback + "_error"]) {
      this[req.callback + "_error"](req, param);
    }
    else {
      this.defaultErrorHandler(req, msg);
    }
  },
  
  defaultErrorHandler: function(req, msg) {
    if (!this.dispatchError(req)) {
      this.notifyStatus("internalError", msg);
      this.setNextTimer();
    }
  },

  dispatchError: function(req) {

    this._hasError = true;

    switch (req.callback) {
    case "direct_messages":
      this.updateTimeline();
      break;
    
    case "statuses_replies":
      this.getDirectMessages();
      break;
    
    case "statuses_friends_timeline":
      this.getReplies();
      break;

    case "account_verify_credentials":
      this.reset();
      break;
    
    default:
      return false;
    }
    return true;
  },

  onError: function(req, param) {

    if (this[req.callback + "_error"]) {
      this[req.callback + "_error"](req, param);
      return;
    }

    this.notifyStatus("internalError", "Network error with " + req.callback);
    if (this.dispatchError(req)) return;

    if (this._timer) {
      this._timer.cancel();
    }
    this.setNextTimer();
  },

  onTimeout: function (req, param) {
    if (this[req.callback + "_timeout"]) {
      this[req.callback + "_timeout"](req, param);
      return;
    }

    this.notifyStatus("internalError", "Request timeout with " + req.callback);
    if (this.dispatchError(req)) return;
    this.setNextTimer();
  },

  rateLimitExceeded: function(req) {
    try {
      var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
      var resp = JSON.decode(req.responseText);
      this.notifyStatus("internalError", resp.error);
    }
    catch (e) {}

    this.get('account.rate_limit_status');
  },

  setInterval: function() {
    var interval = this._rateLimit / 3;
    interval = Math.ceil((60 / interval) * 60 * 1000);
    
    if (this._interval < interval) {
      this._interval = interval;
    }

    // fail safe
    if (this._interval < 180 * 1000) {
      this._interval = 180 * 1000;
    }
  },

  //
  // TwitterAPI callbacks.
  //
  account_verify_credentials: function(obj, req) {
    this._accounts[this._user] = obj;

    this.notifyStatus("updateUsername", obj);
    this.updateTweets();
  },

  account_verify_credentials_error: function(req) {
    this.log("Authentication failed");
    this.notifyStatus("authFail");
    this.reset();
  },

  account_rate_limit_status: function(obj, req) {
    this._rateLimit = obj.hourly_limit;
    this.log("Rate limit is " + this._rateLimit + " requests per hour");
    if (this._rateLimit <= 3) {
      this._rateLimit = 3;
    }

    this.setInterval();

    var nextTime = new Date(obj.reset_time) - Date.now();
    if (nextTime < 180 * 1000) {
      nextTime = 180 * 1000;
    }
    else if (nextTime > 3600 * 1000) {
      nextTime = 3600 * 1000;
    }
    this._timer = this.setDelayTask(nextTime, "updateTweets", null);

    var msg = "Twitter lower rate limit (less than " + this._rateLimit + 
              " requests per hour), request interval has been changed to " + this._interval / 60 / 1000 + " minutes"; 
    this.notifyStatus("showMessage", msg);
    this.log(msg);
    this.log("Auto rate limit control:" + obj.remaining_hits + "/" + obj.hourly_limit + 
             ", reset time: " + obj.reset_time + "(" + Math.ceil(nextTime / 1000 / 60) + " minutes)");
  },

  direct_messages: function(obj, req) {
    this.retrieveTimeline(obj, req, "messages");
    this.updateTimeline();
  },

  statuses_replies: function(obj, req) {
    this.retrieveTimeline(obj, req, "replies");
    this.getDirectMessages();
  },

  statuses_friends_timeline: function(obj, req) {
    this.retrieveTimeline(obj, req, "timeline");
    this.getReplies();
    /*
     * for debugging
       this.updateTimeline();
    */
  },

  statuses_user_timeline: function(obj) {
    try {
      this._accounts[this._user].username = obj[0].user.screen_name;
      this.notifyStatus("updateUsername", obj[0].user.screen_name);
      this.updateTweets();
    }
    catch (e) {
      this.setNextTimer();
    }
  },

  statuses_destroy: function(obj) {
    this.notifyStatus("messageDeleted", {id:obj.id});
    this.deleteStatusFromDB(obj.id, this._accounts[this._user].id);
  },

  statuses_destroy_error: function(req, param) {
    if (Number(req.status) == 404) {
      this.log("Not found the message on server, but delete from local cache anyway");
      this.statuses_destroy(param);
    }
  },

  favorites_destroy: function(obj) {
    this.toggleFavorite(obj.id, false);
    this.notifyStatus("updateFavorite", {id: obj.id, state:false});
  },

  favorites_create: function(obj) {
    this.toggleFavorite(obj.id, true);
    this.notifyStatus("updateFavorite", {id: obj.id, state:true});
  },

  favorites_create_error: function(req) {
  },

  favorites_destroy_error: function(req) {
  },

  toggleFavorite: function(status_id, favorited) {
    var sql = "UPDATE timeline SET favorited = ? WHERE recipient_id = ? AND id = ?";
    var stmt = db.prepare(sql);
    stmt.bindInt32Parameter(0, favorited);
    stmt.bindInt32Parameter(1, this._accounts[this._user].id);
    stmt.bindInt64Parameter(2, status_id);
    try {
      db.begin();
      stmt.execute();
      db.commit();
    }
    catch (e) {
      db.logError("Failed to update favorite on local cache");
      db.rollback();
    }
  },

  statuses_update: function(obj, req) {
    if (obj.id) {
      var status = new Status(obj, "timeline");
      try {
        db.begin();
        status.insertIntoDB(this._accounts[this._user].id);
        db.commit();
      }
      catch (e) {
        db.logError("Failed to insert status");
        db.rollback();
      }
      this.notifyStatus("sentTweet", status);
    }
    else {
      this.notifyStatus("errorOnSendMessage");
    }
  },

  statuses_update_error: function(req) {
    this.reportError("Send error occurred: " + req.status);
    this.log(req.responseText);
    this.notifyStatus("errorOnSendMessage");
  },

  statuses_update_timeout: function() {
    this.reportError("Send timeout occurred");
    this.notifyStatus("errorOnSendMessage");
  },

  direct_messages_new: function(obj, req) {
    if (obj.id) {
      var dm = new DirectMessage(obj);
      try {
        db.begin();
        dm.insertIntoDB(this._accounts[this._user].id);
        db.commit();
      }
      catch (e) {
        db.logError("Failed to insert a direct message");
        db.rollback();
      }
      // Do not call this because TwitterFox doesn't have sent tab
      this.notifyStatus("sentMessage");
    }
    else {
      this.notifyStatus("errorOnSendMessage");
    }
  },

  direct_messages_destroy: function(obj) {
    this.notifyStatus("messageDeleted", {id:obj.id});
    this.deleteMessageFromDB(obj.id, this._accounts[this._user].id);
  },

  direct_messages_destroy_error: function(req, param) {
    if (Number(req.status) == 404) {
      this.log("Not found the message on server, but delete from local cache anyway");
      this.direct_messages_destroy(param);
    }
  },

  direct_messages_new_error: function(req) {
    this.reportError("Send error occurred: " + req.status);
    this.log(req.responseText);
    this.notifyStatus("errorOnSendMessage");
  },

  direct_messages_new_timeout: function() {
    this.reportError("Send timeout occurred");
    this.notifyStatus("errorOnSendMessage");
  },

  // Utilities
  //
  handleCommand: function(data) {
    var msg = eval('(' + data + ')');
    this[msg.command](msg);
  },

  notifyStatus: function(sts, obj) {

    var msg = {"state": sts, "data": obj};

    Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService)
    .notifyObservers(null, "twitternotifier-status", msg.toSource());
  },

  setDelayTask: function(delay, func, data, type) {
    var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer); 

    var target = this;

    if (type == null) {
      type = Ci.nsITimer.TYPE_ONE_SHOT;
    }

    timer.initWithCallback({
      notify: function() {
          target[func](data);
        }
      },
      delay,
      type);
    return timer;
  },

  reportError: function(msg) {
    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.debug")) {
      Components.utils.reportError(msg);
    }
    this.log(msg);
  },

  log: function(msg) {
    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);

    if (pref.getBoolPref("extensions.twitternotifier.debug")) {
      if (this._console == null) 
        this._console = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
      this._console.logStringMessage(msg);
      dump(msg + "\n");
    }
  },

  // for nsISupports
  //
  QueryInterface: function(aIID) {
    // add any other interfaces you support here
    if (!aIID.equals(Ci.nsISupports) && 
        !aIID.equals(Ci.nsIObserver) &&
        !aIID.equals(Ci.nsITwitterFox))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  // for nsIObserver
  //
  observe: function(subject, topic, data) { 
    switch (topic) {

    case "twitternotifier-command":
      this.handleCommand(data);
      break;

    case "xpcom-shutdown":
      this.destroy();
      break;

    }
  },

  // for Debug
  //
  readFile: function(path, filename) {
    var manager = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
    var file = manager.getInstallLocation(TWITTERFOX_UUID).getItemLocation(TWITTERFOX_UUID);
    file.append(path);
    file.append(filename);

    var data = "";
    var fstream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
    var cstream = Cc["@mozilla.org/intl/converter-input-stream;1"].createInstance(Ci.nsIConverterInputStream);
    fstream.init(file, -1, 0, 0);
    cstream.init(fstream, "UTF-8", 0, 0);

    var str = {};
    while (cstream.readString(4096, str) != 0) {
      data += str.value;
    }
    cstream.close();

    return data;
  }
}

//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//
// Singleton
var gTwitterNotifier = null;

// Factory
var TwitterNotifierFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    if (gTwitterNotifier === null) {
      gTwitterNotifier = new TwitterNotifier().QueryInterface(aIID);
    }
    return gTwitterNotifier;
  }
};

// Module
var TwitterNotifierModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);

    Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager)
        .addCategoryEntry("app-startup", 
                          CLASS_NAME,
                          "service," + CONTRACT_ID,
                          true, true);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Ci.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        

    Cc["@mozilla.org/categorymanager;1"]
      .getService(Ci.nsICategoryManager)
        .deleteCategoryEntry("app-startup", 
                             CLASS_NAME,
                             true);
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)  {
    if (!aIID.equals(Ci.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return TwitterNotifierFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

//module initialization
function NSGetModule(aCompMgr, aFileSpec) { return TwitterNotifierModule; }

