//
// Implementation of TwitterFox database client
//
// Copyright (c) 2009 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var Cc = Components.classes;
var Ci = Components.interfaces;
var CLASS_ID = Components.ID("0eef4ea0-ce51-11dd-ad8b-0800200c9a66");
var CLASS_NAME = "TwitterFox Database";
var CONTRACT_ID = "@naan.net/twitterfox-database;1";
var TWITTERFOX_UUID = "twitternotifier@naan.net";
var SQLITE_TEMPLATE = "twitterfox_1.8.sqlite";

function TwitterFoxDatabase()
{
  this._conn = null;
}

TwitterFoxDatabase.prototype = {

  QueryInterface: function(aIID) {
    // add any other interfaces you support here
    if (!aIID.equals(Ci.nsISupports) &&
        !aIID.equals(Ci.nsITwitterFoxDatabase))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  openDatabase: function() {

    var file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    file.append("twitterfox_1.8.sqlite");

    // For debugging
    //
    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);
    if (pref.getBoolPref("extensions.twitternotifier.clearDB")) {
      pref.setBoolPref("extensions.twitternotifier.clearDB", false);
      file.remove(false);
    }

    if (!file.exists()) {
      this.createDatabase();
    }

    try {
      var storageService = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService);
      this._conn = storageService.openDatabase(file);
    }
    catch (e) {
      this.log("Failed to open database: " + e.message);
      return 0;
    }

    // For debugging
    //
    //this.exec("DELETE FROM timeline WHERE id > (SELECT id FROM timeline ORDER BY id DESC LIMIT 1 OFFSET 3);");

    return 1;
  },

  createDatabase: function() {
    var file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    file.append("twitterfox_1.8.sqlite");

    try {
      file.remove(false);
    }
    catch (e) {}

    var manager = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
    var defaultFile = manager.getInstallLocation(TWITTERFOX_UUID).getItemLocation(TWITTERFOX_UUID);
    defaultFile.append("defaults");
    defaultFile.append(SQLITE_TEMPLATE);
    defaultFile.copyTo(file.parent, file.leafName);
    this.log("Copy database file");
  },

  exec: function(sql) {
    try {
      this._conn.executeSimpleSQL(sql);
    }
    catch (e) {
      this.log("Failed to execute query: (" + sql + ") " + this._conn.lastErrorString);
    }
  },

  prepare: function(sql) {
    var stmt;
    try {
      stmt = this._conn.createStatement(sql);
    }
    catch (e) {
      this.log("Failed to prepare statement: (" + sql + ") " + this._conn.lastErrorString);
    }
    return stmt;

  },

  close: function() {
    this._conn.close();
  },

  begin: function() {
    try {
      this._conn.executeSimpleSQL("BEGIN");
    }
    catch (e) {
      this.log("Failed to begin transaction: " + this._conn.lastErrorString);
    }
  },

  commit: function() {
    try {
      this._conn.executeSimpleSQL("COMMIT");
    }
    catch (e) {
      this.log("Failed to commit transaction: " + this._conn.lastErrorString);
    }
  },

  rollback: function() {
    try {
      this._conn.executeSimpleSQL("ROLLBACK");
    }
    catch (e) {
      this.log("Failed to rollback transaction: " + this._conn.lastErrorString);
    }
  },

  insert: function(obj) {
    var sql = "INSERT INTO " + obj._tablename + " (";
    for (var col_name in obj) {
      sql += col_name + ",";
    }
    sql += ") VALUES (";
    for (var value in obj) {
      sql += "?,";
    }
    sql += ")";
    dump(sql);
  },

  logError: function(msg) {

    if (this._conn.lastError >= 100) return;

    msg = msg + " (" + this._conn.lastErrorString + ")";
    this.log(msg);

    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);
    pref.setBoolPref("extensions.twitternotifier.clearDB", true);

    var msg = {"state": "internalError", "data": msg + " Please restart Firefox."};

    Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService).notifyObservers(null, "twitternotifier-status", msg.toSource());
  },
  
  log: function(msg) {
    if (this._console == null) { 
        this._console = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
    }
    this._console.logStringMessage(msg);
    dump(msg + "\n");
  },

  lastErrorString: function() {
    return this._conn.lastErrorString;
  }
};


//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//
// Singleton
var gTwitterFoxDatabase = null;

// Factory
var TwitterFoxDatabaseFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    if (gTwitterFoxDatabase === null) {
      gTwitterFoxDatabase = new TwitterFoxDatabase().QueryInterface(aIID);
    }
    return gTwitterFoxDatabase;
  }
};

// Module
var TwitterFoxDatabaseModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);

    Components.classes["@mozilla.org/categorymanager;1"]
      .getService(Components.interfaces.nsICategoryManager)
        .addCategoryEntry("app-startup", 
                          CLASS_NAME,
                          "service," + CONTRACT_ID,
                          true, true);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        

    Components.classes["@mozilla.org/categorymanager;1"]
      .getService(Components.interfaces.nsICategoryManager)
        .deleteCategoryEntry("app-startup", 
                             CLASS_NAME,
                             true);
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return TwitterFoxDatabaseFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

//module initialization
function NSGetModule(aCompMgr, aFileSpec) { return TwitterFoxDatabaseModule; }

